package gg.malloc.defense.engine;

import java.util.HashSet;
import java.util.HashMap;

import org.bukkit.Location;
import org.bukkit.entity.ArmorStand;
import org.bukkit.entity.Entity;
import org.bukkit.entity.EntityType;
import org.bukkit.GameMode;
import org.bukkit.entity.LivingEntity;
import org.bukkit.entity.Player;
import org.bukkit.entity.Mob;
import org.bukkit.event.entity.EntityDamageEvent;
import org.bukkit.event.entity.EntityDamageByEntityEvent;
import org.bukkit.event.entity.EntityTargetEvent;
import org.bukkit.inventory.EntityEquipment;

import gg.malloc.defense.model.Game;

import gg.malloc.defense.ui.BombCarrier;
import gg.malloc.defense.ui.Items;

public class MobManager {
  LivingEntity m_bombTarget;
  HashSet<Mob> m_livingMobs = new HashSet<>();
  HashSet<Mob> m_bombCarriers = new HashSet<>();
  HashSet<LivingEntity> m_droppedBombs = new HashSet<>();
  HashMap<Mob, Goal> m_mobGoals = new HashMap<>();

  public enum Goal {
    BombCarrier,
    PlayerHarassment
  }

  int m_createdMobs = 0;
  int m_killedMobs = 0;

  public boolean contains(Entity entity) {
    return m_livingMobs.contains(entity);
  }

  public boolean isBombCarrier(Entity entity) {
    return m_bombCarriers.contains(entity);
  }

  public boolean killMob(Entity entity) {
    if (m_livingMobs.contains(entity)) {
      m_killedMobs += 1;
      if (m_bombCarriers.remove(entity)) {
        dropBomb(entity.getLocation());
      }
      return m_livingMobs.remove(entity);
    }
    return false;
  }

  public void clear() {
    for(Mob e : m_livingMobs) {
      e.remove();
    }
    for(LivingEntity e : m_droppedBombs) {
      e.remove();
    }
    if (m_bombTarget != null) {
      m_bombTarget.remove();
    }
    m_bombTarget = null;
    m_droppedBombs.clear();
    m_livingMobs.clear();
    m_bombCarriers.clear();
    m_createdMobs = 0;
    m_killedMobs = 0;
  }

  public int createdMobs() {
    return m_createdMobs;
  }

  public int killedMobs() {
    return m_killedMobs;
  }

  public int remainingMobs() {
    return m_createdMobs - m_killedMobs;
  }

  public double progress() {
    return (double)m_killedMobs / (double)m_createdMobs;
  }

  Player getNearestPlayer(Location location) {
    Player nearestPlayer = null;
    for(Player p : location.getWorld().getPlayers()) {
      if (p.getGameMode() == GameMode.SPECTATOR) {
        continue;
      } else if (nearestPlayer == null) {
        nearestPlayer = p;
      } else if (location.distance(p.getLocation()) < location.distance(nearestPlayer.getLocation())) {
        nearestPlayer = p;
      }
    }
    return nearestPlayer;
  }

  LivingEntity getNearestBombObjective(Location location) {
    LivingEntity nearest = null;
    if (m_droppedBombs.size() > 0) {
      for(LivingEntity ent : m_droppedBombs) {
        if (nearest == null) {
          nearest = ent;
        } else if (location.distance(ent.getLocation()) < location.distance(nearest.getLocation())) {
          nearest = ent;
        }
      }
    } else if (m_bombCarriers.size() > 0) {
      for(LivingEntity ent : m_bombCarriers) {
        if (nearest == null) {
          nearest = ent;
        } else if (location.distance(ent.getLocation()) < location.distance(nearest.getLocation())) {
          nearest = ent;
        }
      }
    }
    if (nearest == null) {
      nearest = m_bombTarget;
    }
    return nearest;
  }
  
  Player getBiggestPlayerThreat(Location location) {
    Entity mostImportantObjective = getNearestBombObjective(location);
    return getNearestPlayer(mostImportantObjective.getLocation());
  }

  LivingEntity getMobTarget(Mob mob) {
    switch(m_mobGoals.get(mob)) {
      case PlayerHarassment:
        if (mob.getTarget() instanceof Player) {
          return mob.getTarget();
        } else {
          Player nearestThreat = getBiggestPlayerThreat(mob.getLocation());
          return nearestThreat;
        }
      case BombCarrier:
        if (m_bombCarriers.contains(mob)) {
          return m_bombTarget;
        } else if (m_droppedBombs.size() > 0) {
          return m_droppedBombs.iterator().next();
        } else {
          return getBiggestPlayerThreat(mob.getLocation());
        }
      default:
        return m_bombTarget;
    }
  }

  public void addEntity(Mob entity, Goal goal) {
    m_livingMobs.add(entity);
    m_mobGoals.put(entity, goal);
    entity.setTarget(getMobTarget(entity));
    m_createdMobs += 1;
  }

  public int bombCount() {
    return m_bombCarriers.size() + m_droppedBombs.size();
  }

  public Mob addBombCarrier(Mob entity) {
    m_bombCarriers.add(entity);
    entity.setTarget(m_bombTarget);
    return (Mob)(new BombCarrier(entity).inHand());
  }

  public void removeBombCarrier(Mob entity) {
    m_bombCarriers.remove(entity);
  }

  public boolean empty() {
    return m_livingMobs.size() == 0;
  }

  public LivingEntity dropBomb(Location location) {
    ArmorStand droppedBomb = (ArmorStand)(new BombCarrier((LivingEntity)location.getWorld().spawnEntity(location, EntityType.ARMOR_STAND)).onHead());
    droppedBomb.setCustomName("The Bomb");
    droppedBomb.setVisible(false);
    droppedBomb.setCustomNameVisible(true);
    droppedBomb.setGlowing(true);
    m_droppedBombs.add(droppedBomb);
    for(Mob ent : m_livingMobs) {
      if (!m_bombCarriers.contains(ent)) {
        ent.setTarget(getMobTarget(ent));
      }
    }
    return droppedBomb;
  }

  boolean m_bombWasHit = false;

  public boolean bombWasHit() {
    return m_bombWasHit;
  }

  public void handleEntityDamage(EntityDamageEvent evt) {
    Entity entity = evt.getEntity();
    m_bombWasHit = false;

    if (entity == m_bombTarget && evt instanceof EntityDamageByEntityEvent) {
      EntityDamageByEntityEvent entityEvt = (EntityDamageByEntityEvent)evt;
      evt.setCancelled(true);
      if (m_bombCarriers.contains(entityEvt.getDamager())) {
        m_bombWasHit = true;
      }
    } if (m_droppedBombs.contains(entity)) {
      evt.setCancelled(true);
      EntityDamageByEntityEvent entityEvt = (EntityDamageByEntityEvent)evt;
      if (m_livingMobs.contains(entityEvt.getDamager())) {
        evt.getEntity().remove();
        m_droppedBombs.remove(evt.getEntity());
        addBombCarrier((Mob)entityEvt.getDamager());
        for(Mob ent : m_livingMobs) {
          ent.setTarget(getMobTarget(ent));
        }
      }
    } else if (m_livingMobs.contains(evt.getEntity())) {
      //m_game.onMobDamaged(evt.getEntity());
    }
  }

  public LivingEntity spawnTarget(Location location) {
    assert(m_bombTarget == null);
    m_bombTarget = (LivingEntity)location.getWorld().spawnEntity(location, EntityType.ARMOR_STAND);
    ArmorStand bombStand = (ArmorStand)m_bombTarget;
    bombStand.setCustomName("Bomb Target");
    bombStand.setVisible(false);
    bombStand.setCustomNameVisible(true);
    bombStand.setGlowing(true);
    EntityEquipment equipment = bombStand.getEquipment();
    equipment.setHelmet(Items.makeBombTarget());
    return m_bombTarget;
  }

  public void handleEntityRetarget(EntityTargetEvent evt) {
    if (m_livingMobs.contains(evt.getEntity())) {
      evt.setTarget(getMobTarget((Mob)evt.getEntity()));
    }
  }
}
